import os

import numpy as np

import iris
import iris.cube
import iris.analysis

import warnings
import datetime

import numba
from numba import vectorize
from numba import cuda
from numba import int32, int64, float32, float64
from numba.cuda.cudadrv import driver

import diagonals

__all__ = ['compute_regsum_2D', 'compute_regsum_3D', 'compute_regsum_levels']


def compute_regsum_2D(var, basins, area):
    """Function that checks device and calls computing functions.

    Checks if the computations are going performed in the CPU or the GPU:

    Parameters
    ----------
    var : float32
        Masked array containing variable data.
    basins : float32
        List containing basin names and masks.
    area : float32
        Masked array containing cell area.

    Returns
    -------
    regmean: float32
        List containing regional mean for variable var
    """
    if diagonals.CONFIG.use_gpu:
        regsum = _compute_regsum_2D_gpu(
            var, basins, area
        )
    else:
        regsum = _compute_regsum_2D_cpu(
            var, basins, area
        )
    return regsum


def compute_regsum_3D(var, basins, volume, tmask):
    """Function that checks device and calls computing functions.

    Checks if the computations are going performed in the CPU or the GPU:

    Parameters
    ----------
    var : float32
        Masked array containing variable data.
    basins : float32
        List containing basin names and masks.
    area : float32
        Masked array containing cell area.

    Returns
    -------
    regmean: float32
        List containing regional mean for variable var
    """
    if diagonals.CONFIG.use_gpu:
        regsum = _compute_regsum_3D_gpu(
            var, basins, volume, tmask
        )
    else:
        regsum = _compute_regsum_3D_cpu(
            var, basins, volume, tmask
        )
    return regsum


def _compute_regsum_2D_cpu(var, basins, area):
    """Function that computes the regional sum for 2D vars in the cpu.

       Computes the weights for each region and performs a weighted sum.

    Parameters
    ----------
    var : float32
        Masked array containing variable data.
    basins : float32
        List containing basin names and masks.
    area : float32
        Masked array containing cell area.

    Returns
    -------
    regmean_total: float32
        List containing regional mean for variable var.
    """
    regsum = {}
    for basin, mask in basins.items():
        weighted_var = _weigh_var_2D(var, mask, area)
        regsum[basin] = np.sum(weighted_var, axis=(1, 2))
    return regsum


def _compute_regsum_3D_cpu(var, basins, volume, tmask):
    """Function that computes the regional sum for 3D vars in the cpu.

    Computes the weights for each region and performs a weighted sum.

    Parameters
    ----------
    var : float32
        Masked array containing variable data.
    basins : float32
        List containing basin names and masks.
    volume : float32
        Masked array containing cell volume

    Returns
    -------
    regmean_total: float32
        List containing regional mean for variable var.
    """
    regsum = {}
    for basin, mask in basins.items():
        weighted_var = _weigh_var_3D(var, mask, volume, tmask)
        regsum[basin] = np.sum(weighted_var, axis=(1, 2, 3))
    return regsum


def compute_regsum_levels(var, basins, volume, tmask):
    """Function that computes the regional sum for 3D vars in the cpu.

    Computes the weights for each region and performs a weighted sum.

    Parameters
    ----------
    var : float32
        Masked array containing variable data.
    basins : float32
        List containing basin names and masks.
    volume : float32
        Masked array containing cell volume
    Returns
    -------
    regmean_total: float32
        List containing regional mean for variable var.
    """
    regsum = {}
    for basin, mask in basins.items():
        weighted_var = _weigh_var_3D(var, mask, volume, tmask)
        regsum[basin] = np.sum(weighted_var, axis=(2, 3))
    return regsum


def _compute_regsum_2D_gpu(var, basins, area):
    """Function that computes the regional sum for 2D vars in the GPU.

    Computes the weights for each region and performs a weighted sum.

    Parameters
    ----------
    var : float32
        Masked array containing variable data.
    basins : float32
        List containing basin names and masks.
    area : float32
        Masked array containing cell area.

    Returns
    -------
    regmean_total: float32
        List containing regional mean for variable var.
    """
    times = var.shape[0]
    lats = var.shape[1]
    lons = var.shape[2]
    mask = len(basins)

    masks = np.empty((mask, lats, lons))
    for i, basin in enumerate(basins):
        masks[i, :, :] = basins[basin]

    block = (128, 1, 1)
    grid_size = (lons // block[0]) + 1
    grid = (grid_size, lats, times)

    gpu_var = cuda.to_device(var.astype(np.float32))
    gpu_area = cuda.to_device(area.astype(np.float32))
    gpu_masks = cuda.to_device(masks)
    gpu_weighted_var = cuda.device_array((mask, times, lats*lons),
                                         dtype=np.float32)

    _variable_weight_cuda[grid, block](gpu_var, gpu_area, gpu_masks,
                                       gpu_weighted_var)

    regsum = {}
    for m in range(mask):
        regsum[m] = {}
        for t in range(times):
            regsum[m][t] = _sum_red_cuda(gpu_weighted_var[m, t, :])

    return regsum


def _compute_regsum_3D_gpu(var, basins, volume, tmask):
    """Function that computes the regional sum for 3D vars in the GPU.

    Computes the weights for each region and performs a weighted sum.

    Parameters
    ----------
    var : float32
        Masked array containing variable data.
    basins : float32
        List containing basin names and masks.
    volume : float32
        Masked array containing cell volume.
    tmask: float32
        Masked array containing land-sea mask for any grid-point type

    Returns
    -------
    regmean_total: float32
        List containing regional mean for variable var.
    """
    times = var.shape[0]
    levs = var.shape[1]
    lats = var.shape[2]
    lons = var.shape[3]
    mask = len(basins)

    masks = np.empty((mask, lats, lons))
    for i, basin in enumerate(basins):
        masks[i, :, :] = basins[basin]

    block = (128, 1, 1)
    grid_size = (lons // block[0]) + 1
    grid = (grid_size, lats, times)

    gpu_var = cuda.to_device(var.astype(np.float32))
    gpu_volume = cuda.to_device(volume.astype(np.float32))
    gpu_tmask = cuda.to_device(tmask.astype(np.float32))
    gpu_masks = cuda.to_device(masks)
    gpu_weighted_var = cuda.device_array((mask, times, levs*lats*lons),
                                         dtype=np.float32)

    _variable_weight_cuda_3D[grid, block](gpu_var, gpu_volume, gpu_tmask,
                                          gpu_masks, gpu_weighted_var)

    regsum = {}
    for m in range(mask):
        regsum[m] = {}
        for t in range(times):
            regsum[m][t] = _sum_red_cuda(gpu_weighted_var[m, t, :])

    return regsum


@vectorize(['float32(float32, float32, float32)'], target='cpu')
def _weigh_var_2D(var, mask, area):
    """Function that weights a 2D variable for each region.

    Parameters
    ----------
    var: float32
        Masked array containing variable to sum.
    mask : float32
        Mask array containing a region mask.
    area: float32
        Masked array containing cell area.

    Returns
    -------
    weights: float32
        Masked array containing the weights for a given region.
    """
    weighted_var = var * mask * area
    return weighted_var


@vectorize(['float32(float32, float32, float32, float32)'],
           target='cpu')
def _weigh_var_3D(var, mask, volume, tmask):
    """Function that weights a 3D variable for each region.

    Parameters
    ----------
    var: float32
        Masked array containing variable to sum.
    mask : float32
        Mask array containing a region mask.
    volume: float32
        Masked array containing cell volume.
    tmask: float32
        Masked array containing land-sea mask for any grid-point type

    Returns
    -------
    weights: float32
        Masked array containing the weights for a given region.
    """
    weighted_var = var * mask * volume * tmask
    return weighted_var


@cuda.jit
def _variable_weight_cuda(var, area, masks, weighted_var):
    """Numba kernel executed on the gpu that weights a 2D variable for each
    region.

    Parameters
    ----------
    var: float32
        Masked array containing variable to sum.
    area: float32
        Masked array containing cell area.
    masks : float32
        Masked array containing region masks.
    weighted_var: float32
        Empty array to store the results.

    Returns
    -------
    weighted_var: float32
        Array containing the weighted variable for a given region.
    """
    i, j, t = cuda.grid(3)
    basins = masks.shape[0]
    x = i+j*var.shape[2]
    if(i >= var.shape[2]):
        return
    for bas in range(basins):
        weighted_var[bas, t, x] = 0.0
        weighted_var[bas, t, x] = var[t, j, i] * area[j, i] * masks[bas, j, i]


@cuda.jit
def _variable_weight_cuda_3D(var, volume, tmask, masks, weighted_var):
    """Numba kernel executed on the gpu that computes the weights and
    weights a 3D variable for each region.

    Parameters
    ----------
    var: float32
        Masked array containing variable to average.
    volume: float32
        Masked array containing cell volume.
    tmask: float32
        Masked array containing land-sea mask for any grid-point type
    masks : float32
        Masked array containing region masks.
    weighted_var: float32
        Empty array to store the results.

    Returns
    -------
    weighted_var: float32
        Array containing the weighted variable for a given region.
    """
    i, j, t = cuda.grid(3)
    basins = masks.shape[0]
    levels = var.shape[1]
    if(i >= var.shape[3]):
        return
    for bas in range(basins):
        for lev in range(levels):
            x = i+var.shape[3]*(j+var.shape[2]*lev)
            weighted_var[bas, t, x] = 0.0
            weighted_var[bas, t, x] = (var[t, j, i] * volume[lev, j, i] *
                                       tmask[lev, j, i] * masks[bas, j, i])


@cuda.reduce
def _sum_red_cuda(x, y):
    """Numba kernel executed on the gpu that performs a sum reduction:

    Parameters
    ----------
    x : float32
        Numpy array.
    y : float32
        Numpy array.

    Returns
    -------
    x+y : float64
        Reduction operation.
    """
    return x+y
